"""
pugh_torch.datasets.__init__

The root dataset path can be set via the environmental variable
``PUGH_TORCH_DATASET_PATH``.

I don't expose this in code because I think it just clutters the code.
"""

import os
from pathlib import Path

ROOT_DATASET_PATH = Path(
    os.environ.get("PUGH_TORCH_DATASET_PATH", "~/.pugh_torch/datasets")
).expanduser()

# Populated automatically via pugh_torch.datasets.Dataset.__init_subclass__
DATASETS = {}

from .base import Dataset
from .torchvision import TorchVisionDataset

import pugh_torch.datasets.classification
import pugh_torch.datasets.segmentation


def get(genre, name):
    """Gets dataset constructor from string identifiers

    Parameters
    ----------
    genre : str
        Type of dataset. e.x. "classification".
        Case insensitive
    name : str
        Name of dataset. e.x. "imagenet".
        Case insensitive
    """

    genre = genre.lower()
    name = name.lower()

    return DATASETS[genre][name]


# Alias for ``get`` function
get_dataset = get



# -*- coding: utf-8 -*-

"""Top-level package for Pugh Torch."""

__author__ = "Brian Pugh"
__email__ = "bnp117@gmail.com"
# Do not edit this string manually, always use bumpversion
# Details in CONTRIBUTING.md
__version__ = "0.3.1"


def get_module_version():
    return __version__


from .exceptions import *

import pugh_torch.datasets
import pugh_torch.helpers
import pugh_torch.losses
import pugh_torch.mappings
import pugh_torch.modules
import pugh_torch.transforms
import pugh_torch.augmentations
import pugh_torch.augmentations as A



from albumentations.core.transforms_interface import DualTransform
import albumentations.augmentations.functional as F

import cv2

__all__ = ["ResizeShortest"]


def _compute_new_shape(length, height, width, trunc=True):
    """Computes the new target shape that results in the shortest side length.

    Parameters
    ----------
    length : float
        Length of the resulting shortest side
    height : float
        Image height
    width : float
        Image width

    Returns
    -------
    int or float
        New height
    int or float
        New width
    """

    if height < width:
        new_height = length
        new_width = width * length / height
    elif width < height:
        new_width = length
        new_height = height * length / width
    else:
        new_width = length
        new_height = length

    if trunc:
        new_height = int(new_height)
        new_width = int(new_width)

    return new_height, new_width


class ResizeShortest(DualTransform):
    """Resize the input so that the shortest side has the given pixel dimension."""

    def __init__(self, length, interpolation=cv2.INTER_LINEAR, always_apply=False, p=1):
        """
        Parameters
        ----------
        length : int
            Length in pixels of the resulting shortest side of the image.
        """

        super().__init__(always_apply, p)
        self.length = length
        self.interpolation = interpolation

    def apply(self, img, interpolation=cv2.INTER_LINEAR, **params):
        new_height, new_width = _compute_new_shape(
            self.length, params["rows"], params["cols"]
        )
        return F.resize(
            img, height=new_height, width=new_width, interpolation=interpolation
        )

    def apply_to_bbox(self, bbox, **params):
        # Bounding box coordinates are scale invariant since coordinates are normalized.
        return bbox

    def apply_to_keypoint(self, keypoint, **params):
        """
        Keypoints are not normalized and need their xy rescaled.
        """

        new_height, new_width = _compute_new_shape(
            self.length, params["rows"], params["cols"]
        )
        scale_x = new_width / params["cols"]
        scale_y = new_height / params["rows"]
        return F.keypoint_scale(keypoint, scale_x, scale_y)

    def get_transform_init_args_names(self):
        # TODO
        return ("length", "interpolation")


# Alias with similar naming to albumentations.augmentations.transforms.LongestMaxSize
ShortestMaxSize = ResizeShortest



"""
Design philosophies/rules:
    * All datasets in this repo are a child of ``Dataset``.
    * All paths are pathlib.Path objects.
        * If something cannot handle it as a Path object, cast it to a string
          as late as possible.
    * Whenever possible, require the least amount of effort on the dev's part
      to get a dataset downloaded and properly formatted.
    * Dataset directories are automatically parsed/derived, so no need to
      prompt the developer on where they want their dataset files.
    * ``self.transform`` is ONLY ever used in the dev's implementation of
      ``self.__getitem__``


To implement your own dataset:
    1. Subclass the ``pugh_torch.datasets.Dataset`` class.
       This class itself is a subclass of ``torch.utils.data.Dataset``.
    2. Implement the download method:
            def download(self):
                # the local folder (guarenteed to exist) is ``self.path``
       This will only be called if the downloaded data isn't available.
       The download being available is determined by a sentinel "downloaded"
       file.
    3. Implement the unpack method:
            def unpack(self):
                # the local folder (guarenteed to exist) is ``self.path``
       This will only be called if the data hasn't been unpacked yet.
       The unpacked being available is determined by a sentinel "unpacked"
       file.

    4. Follow the other remaining instructions at:
        https://pytorch.org/docs/stable/data.html#torch.utils.data.Dataset
"""


import torch
from torchvision import transforms
from . import ROOT_DATASET_PATH, DATASETS


class Dataset(torch.utils.data.Dataset):
    """"""

    def __init_subclass__(cls, **kwargs):
        """Automatic registration stuff"""
        super().__init_subclass__(**kwargs)

        # Register in DATASETS
        modules = cls.__module__.split(".")
        if len(modules) > 3 and modules[0] == "pugh_torch" and modules[1] == "datasets":
            d = DATASETS
            for module in modules[2:-1]:
                if module not in d:
                    d[module] = {}
                d = d[module]
            d[cls.__name__.lower()] = cls

    def __init__(self, split="train", *, transform=None, **kwargs):
        """
        Attempts to download data.

        Parameters
        ----------
        split : str
            One of {"train", "val", "test"}.
            Which data partition to use. Case insensitive.
        transform : obj
            Whatever format you want. Depends on dataset __getitem__ implementation.
            Defaults to just a ``ToTensor`` transform.
            This attribute is NOT used anywhere except in the dataset-specific
            __get__ implementation, or other parent classes of the dataset..
        """

        split = split.lower()
        assert split in ("train", "val", "test")
        self.split = split

        if transform is None:
            self.transform = transforms.Compose(
                [
                    transforms.ToTensor(),
                ]
            )
        else:
            self.transform = transform

        self.path.mkdir(parents=True, exist_ok=True)
        self._download_dataset_if_not_downloaded()
        self._unpack_dataset_if_not_unpacked()

    def _download_dataset_if_not_downloaded(self):
        self.path.mkdir(parents=True, exist_ok=True)

        if self.unpacked or self.downloaded:
            return

        self.download()
        self.downloaded = True

    def _unpack_dataset_if_not_unpacked(self):
        if self.unpacked:
            return

        self.path.mkdir(parents=True, exist_ok=True)

        self.unpack()
        self.unpacked = True

    @property
    def path(self):
        """pathlib.Path to the root of the stored data"""

        try:
            return self.__path
        except AttributeError:
            try:
                dataset_type = self.__class__.__module__.split(".")[
                    -2
                ]  # e.x. "classification", "segmentation"
            except IndexError:
                dataset_type = "unknown"
            self.__path = ROOT_DATASET_PATH / dataset_type / self.__class__.__name__
            return self.__path

    @property
    def downloaded_file(self):
        return self.path / "downloaded"

    @property
    def downloaded(self):
        """We detect if the data has been fully downloaded by a "downloaded"
        file in the root of the data directory.
        """

        return self.downloaded_file.exists()

    @downloaded.setter
    def downloaded(self, val):
        """Touch/Delete sentinel downloaded data file"""

        if val:
            try:
                self.downloaded_file.touch(exist_ok=True)
            except FileNotFoundError as e:
                raise FileNotFoundError(
                    f'Could not create sentinel "{self.downloaded_file}" file, which indicates the directory doesn\'t exist, and so the data most certainly has NOT been downloaded!'
                ) from e
        else:
            self.downloaded_file.unlink(missing_ok=True)

    def download(self):
        """Function to download data to ``self.path``.

        The directories up to ``self.path`` have already been created.

        Will only be called if data has not been downloaded.
        """

        raise NotImplementedError

    @property
    def unpacked_file(self):
        return self.path / "unpacked"

    @property
    def unpacked(self):
        """We detect if the data has been fully unpacked by a "unpacked"
        file in the root of the data directory.
        """

        return self.unpacked_file.exists()

    @unpacked.setter
    def unpacked(self, val):
        """Touch/Delete sentinel unpacked data file"""

        if val:
            try:
                self.unpacked_file.touch(exist_ok=True)
            except FileNotFoundError as e:
                raise FileNotFoundError(
                    f'Could not create sentinel "{self.unpacked_file}" file, which indicates the directory doesn\'t exist, and so the data most certainly has NOT been unpacked!'
                ) from e
        else:
            self.unpacked_file.unlink(missing_ok=True)

    def unpack(self):
        """Post-process the downloaded payload.

        Typically this will be something like unpacking a tar file, or possibly
        re-arranging files.
        """

        raise NotImplementedError



""" Example of a lightly wrapped torchvision dataset.

This is mostly done for greater flexibility of the data processing pipeline.


"""

import torchvision
from torchvision import transforms


from .. import TorchVisionDataset
from ...exceptions import DataUnavailableError
import tarfile


class ImageNet(TorchVisionDataset):
    __download_train_payload = "ILSVRC2012_img_train.tar"
    __download_val_payload = "ILSVRC2012_img_val.tar"
    __download_devkit_payload = "ILSVRC2012_devkit_t12.tar.gz"

    __download_train_link = (
        "https://academictorrents.com/details/a306397ccf9c2ead27155983c254227c0fd938e2"
    )
    __download_val_link = (
        "https://academictorrents.com/details/dfa9ab2528ce76b907047aa8cf8fc792852facb9"
    )
    __download_devkit_link = (
        "https://drive.google.com/file/d/1ina38-4xDAlWVcPcuMZMIx4filKtFqBL"
    )

    def download(self):
        if (self.path / self.__download_train_payload).exists() and (
            self.path / self.__download_val_payload
        ).exists():
            # Definitely don't download if the download payloads already exist.
            return

        raise DataUnavailableError(
            "\n\n"
            "ImageNet 2012 is sort of hard to download; the official website has a bunch of broken links."
            " The easiest way to obtain the data seems to be via:\n"
            f"   train:  {self.__download_train_link}\n"
            f"   val:    {self.__download_val_link}\n"
            f"   devkit: {self.__download_devkit_link}\n"
            "\n"
            "Once downloaded, move the tar files to:\n"
            f"     {self.path / self.__download_train_payload}\n"
            f"     {self.path / self.__download_val_payload}\n"
            f"     {self.path / self.__download_devkit_payload}\n"
        )



import re
import hashlib
from PIL import Image, ImageDraw, ImageFont
import numpy as np


def camel_to_snake(s):
    """Converts a camelCase or pascalCase string to snake_case

    Parameters
    ----------
    s : str
        Some camel/pascal case string to convert to snake case

    Returns
    -------
    str
        camel_case equivalent of the provided string.
    """

    s = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", s)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s).lower()


def compute_sha1(path, chunk_size=2 ** 20):
    """Computes the SHA1 hash of a file

    Parameters
    ----------
    path : str-like
        Path to file to compute the hash of.
    """

    sha1 = hashlib.sha1()
    with open(path, "rb") as f:
        while True:
            chunk = f.read(chunk_size)
            if not chunk:
                break
            sha1.update(chunk)
    return sha1.hexdigest()


def compare_hash(expected, actual):
    """
    Raises
    ------
    pugh_torch.HashMismatchError
        If the hashes do not match.
    """

    if expected == actual:
        return

    raise HashMismatchError(f"{actual} doesn't match expected hash {expected}")


def add_text_under_img(
    img,
    text,
    font_size=None,
    min_font_size=10,
    font="DejaVuSansMono.ttf",
    bg="white",
    fg="black",
):
    """Rasterize and add text under an image.

    Based on:
        https://stackoverflow.com/a/4902713/13376237

    Parameters
    ----------
    img : numpy.ndarray or PIL.Image.Image
        (H,W,3) Image
    text : str
        Text to display under the image
    min_font_size : int
        Minimum font size to render. If the resulting text is wider than the
        passed in img, then the img will be resized.
        Ignored if ``font_size`` is provided.
    font_size : int
        If provided, uses this font size and doesn't auto-search for a font size.
    bg: str or tuple
        Background color of annotation. Defaults to white.
    fg: str or tuple
        Text color. Defaults to black.

    Returns
    -------
    numpy.ndarray
        Resulting annotated image. The image may be rescaled depending on
        ``min_font_size``.
    """

    if isinstance(img, np.ndarray):
        img = Image.fromarray(img)

    if font_size is None:
        # Figure out biggest font that will fit.
        font_size = min_font_size
        font_obj = ImageFont.truetype(font, font_size)
        jump_size = 75
        while True:
            if font_obj.getsize_multiline(text)[0] < img.size[0]:
                font_size += jump_size
            else:
                jump_size = int(jump_size / 2)
                if jump_size == 0 or font_size == min_font_size:
                    break
                font_size -= jump_size
            font_obj = ImageFont.truetype(font, font_size)
            if jump_size <= 1:
                break

        if font_size < min_font_size:
            font_size = min_font_size
            font_obj = ImageFont.truetype(font, font_size)
    else:
        font_obj = ImageFont.truetype(font, font_size)

    # Check and resize img if the font doesn't fit
    font_dim = font_obj.getsize_multiline(text)
    if font_dim[0] > img.size[0]:
        new_w = font_dim[0]
        new_h = int(new_w / img.size[0] * img.size[1])
        img = img.resize((new_w, new_h))

    output = Image.new("RGB", (img.size[0], img.size[1] + font_dim[1]), color=bg)
    output.paste(img)

    draw = ImageDraw.Draw(output)
    draw.text((0, img.size[1]), text, font=font_obj, fill=fg)

    # Convert to a numpy object
    return np.array(output)



""" Lightly wraps torchvision datasets.

This just allows us greater customization without modifying another repo.

Most notably, this:
    * Automatically gets the torchvision dataset constructor based on name
    * Moves the transform responsibility to us
    * Applies our automatic opinionated pathing rules.
"""

import torchvision
from .base import Dataset


class TorchVisionDataset(Dataset):
    auto_construct = True  # Invoke the torchvision dataset constructor in __init__

    def __init_subclass__(cls, **kwargs):
        """Automatically gets the constructor of the torchvision dataset
        of the same name
        """
        super().__init_subclass__(**kwargs)
        cls.torchvision_constructor = getattr(torchvision.datasets, cls.__name__)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if self.auto_construct:
            kwargs["root"] = self.path
            kwargs["transform"] = None  # We handle our own transforms.
            self.dataset = self.torchvision_constructor(**kwargs)

    def __len__(self):
        return len(self.dataset)

    def download(self):
        """ Handled by the torchvision dataset """
        return

    def unpack(self):
        """ Handled by the torchvision dataset """
        return

    @property
    def classes(self):
        return self.dataset.classes

    @property
    def class_to_idx(self):
        return self.dataset.class_to_idx

    def __getitem__(self, index):
        """This is a typical implementation, you can override this with
        your own methods.

        Remember: this is the only location ``self.transform`` is used,
        so its exact type/usage/interpretation is up to you!

        This exact implementation won't work for every dataset.
        """

        img, label = self.dataset[index]
        img = self.transform(img)
        return img, label



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import find_packages, setup

with open("README.md") as f:
    readme = f.read()

with open("CHANGELOG.md") as f:
    changelog = f.read()


setup_requirements = [
    "pytest-runner>=5.2",
]

requirements = [
    "albumentations>=0.4.6",
    "gdown>=3.10.1",
    "numpy>=1.17.1",
    "opencv-python>=4.2.0",
    "scikit-image==0.17.2",
    "tensorboard>=2.2.1",
    "torch>=1.4.0",
    "torchvision>=0.5.0",
]

test_requirements = [
    "black>=19.10b0",
    "codecov>=2.1.4",
    "flake8-debugger>=3.2.1",
    "flake8>=3.8.3",
    "pytest-cov>=2.9.0",
    "pytest-helpers-namespace==2019.1.8",
    "pytest-mock>=3.3.1",
    "pytest-raises>=0.11",
    "pytest>=5.4.3",
]

dev_requirements = [
    "bumpversion>=0.6.0",
    "coverage>=5.1",
    "ipython>=7.15.0",
    "m2r>=0.2.1",
    "pytest-runner>=5.2",
    "Sphinx>=2.0.0b1,<3",
    "sphinx_rtd_theme>=0.4.3",
    "tox>=3.15.2",
    "twine>=3.1.1",
    "wheel>=0.34.2",
]

experiment_requirements = [
    "hydra==2.5",
    "ipdb",
    "matplotlib>=3.0.3",
    "open3d>=0.10.0.0",
    "pytorch-lightning>=0.9.0",
    "scipy>=1.5.1",
    "tqdm>=4.45.0",
]


extra_requirements = {
    "setup": setup_requirements,
    "test": [
        *setup_requirements,
        *requirements,
        *test_requirements,
    ],
    "dev": [
        *setup_requirements,
        *requirements,
        *test_requirements,
        *dev_requirements,
    ],
    "all": [
        *requirements,
        *setup_requirements,
        *test_requirements,
        *dev_requirements,
        *experiment_requirements,
    ],
}

setup(
    author="Brian Pugh",
    author_email="bnp117@gmail.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Functions, losses, and module blocks to share between experiments.",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + changelog,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="pugh_torch",
    name="pugh_torch",
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*"]),
    python_requires=">=3.6",
    setup_requires=setup_requirements,
    test_suite="pugh_torch/tests",
    tests_require=test_requirements,
    extras_require=extra_requirements,
    url="https://github.com/BrianPugh/pugh_torch",
    # Do not edit this string manually, always use bumpversion
    # Details in CONTRIBUTING.rst
    version="0.3.1",
    zip_safe=False,
)



class HashMismatchError(Exception):
    """Occurs when a file does not have the expected hash."""


class DataUnavailableError(Exception):
    """Data is missing from disk and isn't readily available to easily download."""



from .. import TorchVisionDataset


class MNIST(TorchVisionDataset):
    auto_construct = False

    def __init__(self, **kwargs):
        """
        Note both "val" and "test" splits map to "test"
        """

        super().__init__(**kwargs)

        self.dataset = self.torchvision_constructor(
            root=self.path,
            train=self.split == "train",
            transform=None,
            target_transform=None,
            download=True,
        )
